/*
 * Ben NanoNote communication with the Nuelectronics "Color LCD &
 * Joystick Shield" featuring a Nokia 6100 and the PCF8833 display
 * controller.
 *
 * http://shieldlist.org/nuelectronics/colorlcd-joystick
 *
 * Copyright (C) 2013 Paul Boddie
 *
 * SPI functions derived from those in lib/atben.c by Werner Almesberger:
 *
 * Copyright (C) 2010-2011 Werner Almesberger
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#include "pcf8833.h"
#include <stdio.h>
#include <unistd.h>

void LCD_send(lcd_sendmode mode, uint8_t data)
{
    spi_begin();
    spi_send(data, mode);
    spi_end();
}

void LCD_send_more_data(uint8_t data)
{
    spi_begin();
    spi_send(data, LCD_DATA);
    spi_end();
}

void spi_init()
{
    OUT(LCD_CS);
    OUT(LCD_RESET);
    OUT(LCD_CLK);
    OUT(LCD_SEND);
    OUT(LCD_BACKLIGHT);

    CLR(LCD_RESET);
    CLR(LCD_BACKLIGHT);
}

void spi_begin()
{
    CLR(LCD_CS);
}

void spi_end()
{
    SET(LCD_CS);
}

/**
 * Send the given value via MOSI/SEND.
 */
void spi_send(uint8_t v, lcd_sendmode mode)
{
    uint8_t mask;

    if (mode == LCD_DATA)
    {
        #ifdef DEBUG
        printf("D");
        #endif
        SET(LCD_SEND);
    }
    else
    {
        #ifdef DEBUG
        printf("C");
        #endif
        CLR(LCD_SEND);
    }

    SET(LCD_CLK);
    CLR(LCD_CLK);

    for (mask = 0x80; mask; mask >>= 1)
    {
        if (v & mask)
        {
            #ifdef DEBUG
            printf("1");
            #endif
            SET(LCD_SEND);
        }
        else
        {
            #ifdef DEBUG
            printf("0");
            #endif
            CLR(LCD_SEND);
        }

        SET(LCD_CLK);
        CLR(LCD_CLK);
    }

    #ifdef DEBUG
    printf("\n");
    #endif
}

void LCD_init()
{
    /* Perform a hardware reset on the LCD. */
    /* CS=1, CLK=0, D/C=0 apparently superfluous */

    SET(LCD_RESET);
    usleep(50000);
    CLR(LCD_RESET);
    usleep(50000);
    SET(LCD_RESET);
    usleep(50000);

    /* CS=1, CLK=1, D/C=1, 10ms, SWRESET, 10ms apparently superfluous */

    LCD_send(LCD_COMMAND, LCD_SLEEPOUT);

    /* Configure display memory access. */

    LCD_send(LCD_COMMAND, LCD_MADCTL);
    LCD_send(LCD_DATA, LCD_MADCTL_MY_MX);

    /* Contrast setting (suggested value 0x40), 10ms apparently superfluous */

    /* Switch the display on. */

    LCD_send(LCD_COMMAND, LCD_DISPON);

    /* Configure the display mode/format. */
    /* 5 == 16bpp */

    LCD_send(LCD_COMMAND, LCD_COLMOD);
    LCD_send(LCD_DATA, 0x05);

    LCD_send(LCD_COMMAND, LCD_NOP);

    /* Switch the backlight on. */

    SET(LCD_BACKLIGHT);
}

void LCD_off()
{
    /* Switch the backlight off. */

    CLR(LCD_BACKLIGHT);
}

void LCD_window(uint8_t xmin, uint8_t ymin, uint8_t xmax, uint8_t ymax)
{
    LCD_send(LCD_COMMAND, LCD_CASET);
    LCD_send(LCD_DATA, xmin);
    LCD_send(LCD_DATA, xmax);
    LCD_send(LCD_COMMAND, LCD_PASET);
    LCD_send(LCD_DATA, ymin);
    LCD_send(LCD_DATA, ymax);
}

void LCD_blit_int(uint16_t colour)
{
    LCD_send_more_data((colour >> 8) & 0xff);                              /* RGGB -> RG */
    LCD_send_more_data(colour & 0xff);                                     /* RGGB -> GB */
}

void LCD_area(uint8_t xmin, uint8_t ymin, uint8_t xmax, uint8_t ymax, uint16_t colour)
{
    uint16_t i, limit = ((xmax - xmin + 1) * (ymax - ymin + 1)) + 1;

    LCD_window(xmin, ymin, xmax, ymax);
    LCD_send(LCD_COMMAND, LCD_RAMWR);

    for (i = 0; i < limit; i++)
    {
        LCD_blit_int(colour);
    }

    LCD_send(LCD_COMMAND, LCD_NOP);
}

void LCD_image(int x, int y, const uint16_t image[], uint8_t width, uint8_t height)
{
    LCD_image_region(x, y, image, width, height, 0, width, 0, height);
}

void LCD_image_region(int x, int y, const uint16_t image[], uint8_t width, uint8_t height, uint8_t from_x, uint8_t span_x, uint8_t from_y, uint8_t span_y)
{
    uint8_t xmin, ymin, xcmax, ycmax, xcmin, xc, yc;
    uint16_t xmax, ymax, colour;

    /* Find the screen ranges for blitting. */

    xmin = (x < 0) ? 0 : x;
    ymin = (y < 0) ? 0 : y;
    xmax = (x + span_x > SCREEN_X_MAX) ? SCREEN_X_MAX : x + span_x - 1;
    ymax = (y + span_y > SCREEN_Y_MAX) ? SCREEN_Y_MAX : y + span_y - 1;

    if ((xmin > SCREEN_X_MAX) || (ymin > SCREEN_Y_MAX) || (xmax < 0) || (ymax < 0))
    {
        return;
    }

    /* Find the image ranges for reading data. */

    xcmin = (x < 0) ? -x + from_x : from_x;
    yc = (y < 0) ? -y + from_y : from_y;
    xcmax = from_x + ((x + span_x > SCREEN_WIDTH) ? SCREEN_WIDTH - x : span_x);
    ycmax = from_y + ((y + span_y > SCREEN_HEIGHT) ? SCREEN_HEIGHT - y : span_y);

    LCD_window(xmin, ymin, xmax, ymax);
    LCD_send(LCD_COMMAND, LCD_RAMWR);

    /* Reading from the appropriate image ranges, blit to the appropriate
       screen ranges. */

    xc = xcmin;
    while (yc < ycmax)
    {
        colour = image[yc * width + xc];
        LCD_blit_int(colour);

        xc++;
        if (xc == xcmax)
        {
            xc = xcmin;
            yc++;
        }
    }

    LCD_send(LCD_COMMAND, LCD_NOP);
}

void LCD_normal()
{
    LCD_send(LCD_COMMAND, LCD_NORON);
}

void LCD_scroll_area(uint8_t top_fixed, uint8_t scrolling, uint8_t bottom_fixed)
{
    LCD_send(LCD_COMMAND, LCD_VSCRDEF);
    LCD_send(LCD_DATA, top_fixed);
    LCD_send(LCD_DATA, scrolling);
    LCD_send(LCD_DATA, bottom_fixed);
}

void LCD_scroll_start(uint8_t address)
{
    LCD_send(LCD_COMMAND, LCD_SEP);
    LCD_send(LCD_DATA, address);
}

/* vim: tabstop=4 expandtab shiftwidth=4
 */
